@@ -1,6 +1,6 @@
 /***************************************************************************
  *   fheroes2: https://github.com/ihhub/fheroes2                           *
- *   Copyright (C) 2019 - 2023                                             *
+ *   Copyright (C) 2019 - 2025                                             *
  *                                                                         *
  *   Free Heroes2 Engine: http://sourceforge.net/projects/fheroes2         *
  *   Copyright (C) 2009 by Andrey Afletdinov <fheroes2@gmail.com>          *
@@ -22,6 +22,7 @@
  ***************************************************************************/
 
 #include <algorithm>
+#include <cassert>
 #include <cstdint>
 #include <ctime>
 #include <iomanip>
@@ -35,9 +36,8 @@
 
 #include "agg_image.h"
 #include "cursor.h"
-#include "dialog.h"
+#include "dialog.h" // IWYU pragma: associated
 #include "dir.h"
-#include "game_delays.h"
 #include "game_hotkeys.h"
 #include "game_io.h"
 #include "icn.h"
@@ -49,7 +49,6 @@
 #include "screen.h"
 #include "settings.h"
 #include "system.h"
-#include "tools.h"
 #include "translations.h"
 #include "ui_button.h"
 #include "ui_dialog.h"
@@ -65,16 +64,6 @@ namespace
     // This constant sets the maximum displayed file name width. This value affects the dialog horizontal size.
     const int32_t maxFileNameWidth = 260;
 
-    std::string ResizeToShortName( const std::string & str )
-    {
-        std::string res = System::GetBasename( str );
-        const size_t it = res.rfind( '.' );
-        if ( std::string::npos != it ) {
-            res.resize( it );
-        }
-        return res;
-    }
-
     void redrawDateTime( fheroes2::Image & output, const time_t timestamp, const int32_t dstx, const int32_t dsty, const fheroes2::FontType font )
     {
         const size_t arraySize = 5;
@@ -111,20 +100,19 @@ namespace
         text.draw( dstx + 105 - text.width() / 2, dsty, output );
     }
 
-    bool redrawTextInputField( const std::string & filename, const fheroes2::Rect & field, const bool isEditing )
+    void redrawTextInputField( const std::string & filename, const fheroes2::Rect & field, fheroes2::Image & output )
     {
         if ( filename.empty() ) {
-            return false;
+            return;
         }
 
-        fheroes2::Display & display = fheroes2::Display::instance();
+        fheroes2::Text textInput( filename, fheroes2::FontType::normalYellow() );
 
-        fheroes2::Text currentFilename( filename, isEditing ? fheroes2::FontType::normalWhite() : fheroes2::FontType::normalYellow() );
-        const int32_t initialTextWidth = currentFilename.width();
-        currentFilename.fitToOneRow( maxFileNameWidth );
-        currentFilename.draw( field.x + 4 + ( maxFileNameWidth - currentFilename.width() ) / 2, field.y + 4, display );
+        // Do not ignore spaces at the end.
+        textInput.keepLineTrailingSpaces();
+        textInput.fitToOneRow( maxFileNameWidth );
 
-        return ( initialTextWidth + 10 ) > maxFileNameWidth;
+        textInput.drawInRoi( field.x + 4 + ( maxFileNameWidth - textInput.width() ) / 2, field.y + 4, output, field );
     }
 
     class FileInfoListBox : public Interface::ListBox<Maps::FileInfo>
@@ -146,19 +134,12 @@ namespace
 
         void ActionListPressRight( Maps::FileInfo & info ) override
         {
-            // On some OSes like Windows, the path may contain '\' symbols. This symbol doesn't exist in the resources.
-            // To avoid this we have to replace all '\' symbols by '/' symbols.
-            std::string fullPath = info.file;
-
-            // TODO: Make '\' symbol in small font to properly show file path in OS familiar style.
-            StringReplace( fullPath, "\\", "/" );
-
-            const fheroes2::Text header( ResizeToShortName( info.file ), fheroes2::FontType::normalYellow() );
+            const fheroes2::Text header( System::GetStem( info.filename ), fheroes2::FontType::normalYellow() );
 
             fheroes2::MultiFontText body;
 
             body.add( { _( "Map: " ), fheroes2::FontType::normalYellow() } );
-            body.add( { info.name, fheroes2::FontType::normalWhite() } );
+            body.add( { info.name, fheroes2::FontType::normalWhite(), info.getSupportedLanguage() } );
 
             if ( info.worldDay > 0 || info.worldWeek > 0 || info.worldMonth > 0 ) {
                 body.add( { _( "\n\nMonth: " ), fheroes2::FontType::normalYellow() } );
@@ -170,7 +151,7 @@ namespace
             }
 
             body.add( { _( "\n\nLocation: " ), fheroes2::FontType::smallYellow() } );
-            body.add( { fullPath, fheroes2::FontType::smallWhite() } );
+            body.add( { info.filename, fheroes2::FontType::smallWhite() } );
 
             fheroes2::showMessage( header, body, Dialog::ZERO );
         }
@@ -206,18 +187,8 @@ namespace
 
     void FileInfoListBox::RedrawItem( const Maps::FileInfo & info, int32_t dstx, int32_t dsty, bool current )
     {
-        std::string savname( System::GetBasename( info.file ) );
-
-        if ( savname.empty() ) {
-            return;
-        }
-
-        const std::string saveExtension = Game::GetSaveFileExtension();
-        const size_t dotPos = savname.size() - saveExtension.size();
-
-        if ( StringLower( savname.substr( dotPos ) ) == saveExtension ) {
-            savname.erase( dotPos );
-        }
+        std::string savname( System::GetStem( info.filename ) );
+        assert( !savname.empty() );
 
         const fheroes2::FontType font = current ? fheroes2::FontType::normalYellow() : fheroes2::FontType::normalWhite();
         fheroes2::Display & display = fheroes2::Display::instance();
@@ -225,6 +196,7 @@ namespace
         dsty += 2;
 
         fheroes2::Text text{ std::move( savname ), font };
+        text.keepLineTrailingSpaces();
         text.fitToOneRow( maxFileNameWidth );
         text.draw( dstx + 4 + ( maxFileNameWidth - text.width() ) / 2, dsty, display );
 
@@ -258,22 +230,23 @@ namespace
 
     MapsFileInfoList getSortedMapsFileInfoList()
     {
-        ListFiles list1;
-        list1.ReadDir( Game::GetSaveDir(), Game::GetSaveFileExtension(), false );
+        ListFiles files;
+        files.ReadDir( Game::GetSaveDir(), Game::GetSaveFileExtension() );
 
-        MapsFileInfoList list2( list1.size() );
-        size_t saveFileCount = 0;
-        for ( const std::string & saveFile : list1 ) {
-            if ( list2[saveFileCount].ReadSAV( saveFile ) ) {
-                ++saveFileCount;
+        MapsFileInfoList mapInfos;
+        mapInfos.reserve( files.size() );
+
+        for ( std::string & saveFile : files ) {
+            Maps::FileInfo mapInfo;
+
+            if ( Game::LoadSAV2FileInfo( std::move( saveFile ), mapInfo ) ) {
+                mapInfos.emplace_back( std::move( mapInfo ) );
             }
         }
-        if ( saveFileCount != list2.size() ) {
-            list2.resize( saveFileCount );
-        }
-        std::sort( list2.begin(), list2.end(), Maps::FileInfo::FileSorting );
 
-        return list2;
+        std::sort( mapInfos.begin(), mapInfos.end(), Maps::FileInfo::sortByFileName );
+
+        return mapInfos;
     }
 
     std::string selectFileListSimple( const std::string & header, const std::string & lastfile, const bool isEditing )
@@ -298,8 +271,8 @@ namespace
         // Dialog height is also capped with the current screen height.
         fheroes2::StandardWindow background( maxFileNameWidth + 204, std::min( display.height() - 100, maxDialogHeight ), true, display );
 
-        const fheroes2::Rect area = background.activeArea();
-        const fheroes2::Rect listRoi( area.x + 24, area.y + 37, area.width - 75, area.height - listHeightDeduction );
+        const fheroes2::Rect dialogArea( background.activeArea() );
+        const fheroes2::Rect listRoi( dialogArea.x + 24, dialogArea.y + 37, dialogArea.width - 75, dialogArea.height - listHeightDeduction );
         const fheroes2::Rect textInputRoi( listRoi.x, listRoi.y + listRoi.height + 12, maxFileNameWidth + 8, 21 );
         const int32_t dateTimeoffsetX = textInputRoi.x + textInputRoi.width;
         const int32_t dateTimeWidth = listRoi.width - textInputRoi.width;
@@ -311,71 +284,35 @@ namespace
         // Make background for the selected file date and time.
         background.applyTextBackgroundShading( { dateTimeoffsetX, textInputRoi.y, dateTimeWidth, textInputRoi.height } );
 
-        fheroes2::ImageRestorer textInputAndDateBackground( fheroes2::Display::instance(), textInputRoi.x, textInputRoi.y, listRoi.width, textInputRoi.height );
-
-        const bool isEvilInterface = Settings::Get().isEvilInterfaceEnabled();
+        fheroes2::ImageRestorer textInputBackground( display, textInputRoi.x, textInputRoi.y, textInputRoi.width, textInputRoi.height );
+        fheroes2::ImageRestorer dateBackground( display, dateTimeoffsetX, textInputRoi.y, dateTimeWidth, textInputRoi.height );
+        const fheroes2::Rect textInputAndDateROI( textInputRoi.x, textInputRoi.y, listRoi.width, textInputRoi.height );
 
         // Prepare OKAY and CANCEL buttons and render their shadows.
-        const int buttonOkIcn = isEvilInterface ? ICN::BUTTON_SMALL_OKAY_EVIL : ICN::BUTTON_SMALL_OKAY_GOOD;
-        const int32_t buttonOffsetY = area.y + area.height - 32;
-        fheroes2::Button buttonOk( area.x + 18, buttonOffsetY, buttonOkIcn, 0, 1 );
-        fheroes2::addGradientShadow( fheroes2::AGG::GetICN( buttonOkIcn, 0 ), display, buttonOk.area().getPosition(), { -5, 5 } );
+        fheroes2::Button buttonOk;
         if ( !isEditing && lists.empty() ) {
             buttonOk.disable();
         }
-        buttonOk.draw();
+        fheroes2::Button buttonCancel;
 
-        const int buttonCancelIcn = isEvilInterface ? ICN::BUTTON_SMALL_CANCEL_EVIL : ICN::BUTTON_SMALL_CANCEL_GOOD;
-        const fheroes2::Sprite & buttonCancelSprite = fheroes2::AGG::GetICN( buttonCancelIcn, 0 );
-        fheroes2::Button buttonCancel( area.x + area.width - buttonCancelSprite.width() - 18, buttonOffsetY, buttonCancelIcn, 0, 1 );
-        fheroes2::addGradientShadow( buttonCancelSprite, display, buttonCancel.area().getPosition(), { -5, 5 } );
-        buttonCancel.draw();
+        background.renderOkayCancelButtons( buttonOk, buttonCancel );
 
-        // Virtual keyboard button is used only in save game mode (when 'isEditing' is true ).
-        std::unique_ptr<fheroes2::ButtonSprite> buttonVirtualKB;
-
-        FileInfoListBox listbox( area.getPosition() );
+        FileInfoListBox listbox( dialogArea.getPosition() );
 
         // Initialize list background restorer to use it in list method 'listbox.RedrawBackground()'.
         listbox.initListBackgroundRestorer( listRoi );
 
         listbox.SetAreaItems( { listRoi.x, listRoi.y + 3, listRoi.width - listAreaOffsetY, listRoi.height - listAreaHeightDeduction } );
 
-        // Render the scrollbar.
-        const fheroes2::Sprite & scrollBar = fheroes2::AGG::GetICN( isEvilInterface ? ICN::ADVBORDE : ICN::ADVBORD, 0 );
-        int32_t scrollbarOffsetX = area.x + area.width - 35;
+        const bool isEvilInterface = Settings::Get().isEvilInterfaceEnabled();
 
-        // Top part of scrollbar background.
-        const int32_t topPartHeight = 19;
-        const int32_t scrollBarWidth = 16;
-        fheroes2::Copy( scrollBar, 536, 176, display, scrollbarOffsetX, listRoi.y, scrollBarWidth, topPartHeight );
-
-        // Middle part of scrollbar background.
-        int32_t offsetY = topPartHeight;
-        const int32_t middlePartHeight = 88;
-        const int32_t middlePartCount = ( listRoi.height - 2 * topPartHeight + middlePartHeight - 1 ) / middlePartHeight;
-
-        for ( int32_t i = 0; i < middlePartCount; ++i ) {
-            fheroes2::Copy( scrollBar, 536, 196, display, scrollbarOffsetX, listRoi.y + offsetY, scrollBarWidth,
-                            std::min( middlePartHeight, listRoi.height - offsetY - topPartHeight ) );
-            offsetY += middlePartHeight;
-        }
-
-        // Bottom part of scrollbar background.
-        fheroes2::Copy( scrollBar, 536, 285, display, scrollbarOffsetX, listRoi.y + listRoi.height - topPartHeight, scrollBarWidth, topPartHeight );
+        int32_t scrollbarOffsetX = dialogArea.x + dialogArea.width - 35;
+        background.renderScrollbarBackground( { scrollbarOffsetX, listRoi.y, listRoi.width, listRoi.height }, isEvilInterface );
 
         const int listIcnId = isEvilInterface ? ICN::SCROLLE : ICN::SCROLL;
-
+        const int32_t topPartHeight = 19;
         ++scrollbarOffsetX;
 
-        // Make scrollbar shadow.
-        for ( uint8_t i = 0; i < 4; ++i ) {
-            const uint8_t transformId = i + 2;
-            const int32_t sizeCorrection = i + 1;
-            fheroes2::ApplyTransform( display, scrollbarOffsetX - transformId, listRoi.y + sizeCorrection, 1, listRoi.height - sizeCorrection, transformId );
-            fheroes2::ApplyTransform( display, scrollbarOffsetX - transformId, listRoi.y + listRoi.height + i, scrollBarWidth, 1, transformId );
-        }
-
         listbox.SetScrollButtonUp( listIcnId, 0, 1, { scrollbarOffsetX, listRoi.y + 1 } );
         listbox.SetScrollButtonDn( listIcnId, 2, 3, { scrollbarOffsetX, listRoi.y + listRoi.height - 15 } );
         listbox.setScrollBarArea( { scrollbarOffsetX + 2, listRoi.y + topPartHeight, 10, listRoi.height - 2 * topPartHeight } );
@@ -388,12 +325,12 @@ namespace
         size_t charInsertPos = 0;
 
         if ( !lastfile.empty() ) {
-            filename = ResizeToShortName( lastfile );
+            filename = System::GetStem( lastfile );
             charInsertPos = filename.size();
 
             MapsFileInfoList::iterator it = lists.begin();
             for ( ; it != lists.end(); ++it ) {
-                if ( ( *it ).file == lastfile ) {
+                if ( ( *it ).filename == lastfile ) {
                     break;
                 }
             }
@@ -411,53 +348,71 @@ namespace
         }
 
         if ( filename.empty() && listbox.isSelected() ) {
-            filename = ResizeToShortName( listbox.GetCurrent().file );
+            filename = System::GetStem( listbox.GetCurrent().filename );
             charInsertPos = filename.size();
         }
 
+        auto buttonOkDisabler = [&buttonOk, &filename]() {
+            if ( filename.empty() && buttonOk.isEnabled() ) {
+                buttonOk.disable();
+                buttonOk.draw();
+            }
+            else if ( !filename.empty() && buttonOk.isDisabled() ) {
+                buttonOk.enable();
+                buttonOk.draw();
+            }
+        };
+
         listbox.Redraw();
-        redrawTextInputField( filename, textInputRoi, isEditing );
+
+        // Virtual keyboard button, text input and blinking cursor are used only in save game mode (when 'isEditing' is true ).
+        std::unique_ptr<fheroes2::ButtonSprite> buttonVirtualKB;
+        std::unique_ptr<fheroes2::TextInputField> textInput;
 
         const fheroes2::Text title( header, fheroes2::FontType::normalYellow() );
-        title.draw( area.x + ( area.width - title.width() ) / 2, area.y + 16, display );
+        title.drawInRoi( dialogArea.x + ( dialogArea.width - title.width() ) / 2, dialogArea.y + 16, display, dialogArea );
 
         if ( isEditing ) {
-            // Generate and render a button to open the Virtual Keyboard window.
-            fheroes2::Sprite released;
-            fheroes2::Sprite pressed;
+            // Render a button to open the Virtual Keyboard window.
+            buttonVirtualKB = std::make_unique<fheroes2::ButtonSprite>();
+            background.renderCustomButtonSprite( *buttonVirtualKB, "...", { 48, 25 }, { 0, 7 }, fheroes2::StandardWindow::Padding::BOTTOM_CENTER );
 
-            const int32_t buttonWidth = buttonCancelSprite.width() / 2;
+            // Prepare the text input and set it to always fit the file name input field width.
+            textInput = std::make_unique<fheroes2::TextInputField>( fheroes2::Rect{ textInputRoi.x + 4, textInputRoi.y + 2, maxFileNameWidth, textInputRoi.height - 2 },
+                                                                    false, true, display );
 
-            makeButtonSprites( released, pressed, "...", buttonWidth, isEvilInterface, false );
-            buttonVirtualKB = std::make_unique<fheroes2::ButtonSprite>( area.x + ( area.width - buttonWidth ) / 2, buttonOffsetY, released, pressed );
-
-            fheroes2::addGradientShadow( released, display, buttonVirtualKB->area().getPosition(), { -5, 5 } );
-            buttonVirtualKB->draw();
-
-            Game::passAnimationDelay( Game::DelayType::CURSOR_BLINK_DELAY );
+            if ( !listbox.isSelected() ) {
+                textInput->draw( filename, static_cast<int32_t>( charInsertPos ) );
+                redrawDateTime( display, std::time( nullptr ), dateTimeoffsetX, textInputRoi.y + 4, fheroes2::FontType::normalWhite() );
+            }
         }
-        else if ( listbox.isSelected() ) {
-            // Render the saved file date and time.
+
+        if ( listbox.isSelected() ) {
+            // Render the saved file name, date and time.
+            redrawTextInputField( filename, textInputRoi, display );
             redrawDateTime( display, listbox.GetCurrent().timestamp, dateTimeoffsetX, textInputRoi.y + 4, fheroes2::FontType::normalYellow() );
         }
 
         display.render( background.totalArea() );
 
         std::string result;
-        bool isTextLimit = false;
         std::string lastSelectedSaveFileName;
 
         const bool isInGameKeyboardRequired = System::isVirtualKeyboardSupported();
 
-        bool isCursorVisible = true;
+        const size_t lengthLimit{ 255 };
 
         LocalEvent & le = LocalEvent::Get();
 
-        while ( le.HandleEvents( !isEditing || Game::isDelayNeeded( { Game::DelayType::CURSOR_BLINK_DELAY } ) ) && result.empty() ) {
-            le.MousePressLeft( buttonOk.area() ) && buttonOk.isEnabled() ? buttonOk.drawOnPress() : buttonOk.drawOnRelease();
-            le.MousePressLeft( buttonCancel.area() ) ? buttonCancel.drawOnPress() : buttonCancel.drawOnRelease();
+        while ( le.HandleEvents() && result.empty() ) {
+            buttonOk.drawOnState( le.isMouseLeftButtonPressedInArea( buttonOk.area() ) );
+            buttonCancel.drawOnState( le.isMouseLeftButtonPressedInArea( buttonCancel.area() ) );
             if ( isEditing ) {
-                le.MousePressLeft( buttonVirtualKB->area() ) ? buttonVirtualKB->drawOnPress() : buttonVirtualKB->drawOnRelease();
+                buttonVirtualKB->drawOnState( le.isMouseLeftButtonPressedInArea( buttonVirtualKB->area() ) );
+            }
+
+            if ( le.MouseClickLeft( buttonCancel.area() ) || Game::HotKeyPressEvent( Game::HotKeyEvent::DEFAULT_CANCEL ) ) {
+                return {};
             }
 
             const int listId = listbox.getCurrentId();
@@ -466,54 +421,85 @@ namespace
 
             bool isListboxSelected = listbox.isSelected();
 
-            bool needRedraw = listId != listbox.getCurrentId();
+            bool needRedraw = ( listId != listbox.getCurrentId() );
 
-            if ( ( buttonOk.isEnabled() && le.MouseClickLeft( buttonOk.area() ) ) || Game::HotKeyPressEvent( Game::HotKeyEvent::DEFAULT_OKAY )
-                 || listbox.isDoubleClicked() ) {
-                if ( !filename.empty() ) {
+            if ( le.isKeyPressed( fheroes2::Key::KEY_DELETE ) && isListboxSelected ) {
+                listbox.SetCurrent( listId );
+                listbox.Redraw();
+
+                std::string msg( _( "Are you sure you want to delete file:" ) );
+                msg.append( "\n\n" );
+                msg.append( System::GetFileName( listbox.GetCurrent().filename ) );
+
+                if ( Dialog::YES == fheroes2::showStandardTextMessage( _( "Warning" ), msg, Dialog::YES | Dialog::NO ) ) {
+                    System::Unlink( listbox.GetCurrent().filename );
+                    listbox.RemoveSelected();
+
+                    if ( lists.empty() ) {
+                        listbox.Redraw();
+
+                        isListboxSelected = false;
+                        charInsertPos = 0;
+                        filename.clear();
+
+                        buttonOk.disable();
+                        buttonOk.draw();
+                    }
+
+                    listbox.updateScrollBarImage();
+                    listbox.SetCurrent( std::max( listId - 1, 0 ) );
+                }
+
+                needRedraw = true;
+            }
+            else if ( ( buttonOk.isEnabled() && le.MouseClickLeft( buttonOk.area() ) ) || Game::HotKeyPressEvent( Game::HotKeyEvent::DEFAULT_OKAY )
+                      || listbox.isDoubleClicked() ) {
+                if ( isListboxSelected ) {
+                    result = listbox.GetCurrent().filename;
+                }
+                else if ( !filename.empty() ) {
                     result = System::concatPath( Game::GetSaveDir(), filename + Game::GetSaveFileExtension() );
                 }
-                else if ( isListboxSelected ) {
-                    result = listbox.GetCurrent().file;
-                }
-            }
-            else if ( le.MouseClickLeft( buttonCancel.area() ) || Game::HotKeyPressEvent( Game::HotKeyEvent::DEFAULT_CANCEL ) ) {
-                break;
             }
             else if ( isEditing ) {
+                assert( textInput != nullptr );
+
+                bool prepareRedraw = false;
+
                 if ( le.MouseClickLeft( buttonVirtualKB->area() ) || ( isInGameKeyboardRequired && le.MouseClickLeft( textInputRoi ) ) ) {
-                    fheroes2::openVirtualKeyboard( filename );
+                    fheroes2::openVirtualKeyboard( filename, lengthLimit );
 
                     charInsertPos = filename.size();
-                    listbox.Unselect();
-                    isListboxSelected = false;
-                    needRedraw = true;
+                    prepareRedraw = true;
 
                     // Set the whole screen to redraw next time to properly restore image under the Virtual Keyboard dialog.
                     display.updateNextRenderRoi( { 0, 0, display.width(), display.height() } );
                 }
-                else if ( le.MouseClickLeft( textInputRoi ) ) {
-                    const fheroes2::Text text( filename, fheroes2::FontType::normalWhite() );
-                    const int32_t textStartOffsetX = isTextLimit ? 8 : ( textInputRoi.width - text.width() ) / 2;
-                    charInsertPos = fheroes2::getTextInputCursorPosition( filename, fheroes2::FontType::normalWhite(), charInsertPos, le.GetMouseCursor().x,
-                                                                          textInputRoi.x + textStartOffsetX );
-                    if ( filename.empty() ) {
-                        buttonOk.disable();
+                else if ( !filename.empty() && le.MouseClickLeft( textInputRoi ) ) {
+                    if ( isListboxSelected ) {
+                        // To proper position the text cursor update the textInput text and set cursor position to 0 to correspond
+                        // the behavior of `fitToOneRow` of `fheroes2::Text` before calling `fheroes2::getTextInputCursorPosition()`.
+                        textInput->set( filename, 0 );
                     }
 
-                    listbox.Unselect();
-                    isListboxSelected = false;
-                    needRedraw = true;
+                    const size_t newPos = textInput->getCursorInTextPosition( le.getMouseLeftButtonPressedPos() );
+
+                    if ( newPos != charInsertPos || isListboxSelected ) {
+                        charInsertPos = newPos;
+                        prepareRedraw = true;
+                    }
                 }
-                else if ( !listboxEvent && le.KeyPress()
-                          && ( !isTextLimit || fheroes2::Key::KEY_BACKSPACE == le.KeyValue() || fheroes2::Key::KEY_DELETE == le.KeyValue() ) ) {
-                    charInsertPos = InsertKeySym( filename, charInsertPos, le.KeyValue(), LocalEvent::getCurrentKeyModifiers() );
-                    if ( filename.empty() ) {
-                        buttonOk.disable();
-                    }
-                    else {
-                        buttonOk.enable();
+                else if ( !listboxEvent && le.isAnyKeyPressed() ) {
+                    const fheroes2::Key pressedKey = le.getPressedKeyValue();
+                    if ( ( filename.size() < lengthLimit || pressedKey == fheroes2::Key::KEY_BACKSPACE || pressedKey == fheroes2::Key::KEY_DELETE )
+                         && pressedKey != fheroes2::Key::KEY_UP && pressedKey != fheroes2::Key::KEY_DOWN ) {
+                        charInsertPos = InsertKeySym( filename, charInsertPos, pressedKey, LocalEvent::getCurrentKeyModifiers() );
+                        prepareRedraw = true;
                     }
+                }
+
+                if ( prepareRedraw ) {
+                    buttonOkDisabler();
 
                     needRedraw = true;
                     listbox.Unselect();
@@ -521,10 +507,10 @@ namespace
                 }
             }
 
-            if ( le.MousePressRight( buttonCancel.area() ) ) {
+            if ( le.isMouseRightButtonPressedInArea( buttonCancel.area() ) ) {
                 fheroes2::showStandardTextMessage( _( "Cancel" ), _( "Exit this menu without doing anything." ), Dialog::ZERO );
             }
-            else if ( le.MousePressRight( buttonOk.area() ) ) {
+            else if ( le.isMouseRightButtonPressedInArea( buttonOk.area() ) ) {
                 if ( isEditing ) {
                     fheroes2::showStandardTextMessage( _( "Okay" ), _( "Click to save the current game." ), Dialog::ZERO );
                 }
@@ -532,71 +518,57 @@ namespace
                     fheroes2::showStandardTextMessage( _( "Okay" ), _( "Click to load a previously saved game." ), Dialog::ZERO );
                 }
             }
-            else if ( isEditing && le.MousePressRight( buttonVirtualKB->area() ) ) {
+            else if ( isEditing && le.isMouseRightButtonPressedInArea( buttonVirtualKB->area() ) ) {
                 fheroes2::showStandardTextMessage( _( "Open Virtual Keyboard" ), _( "Click to open the Virtual Keyboard dialog." ), Dialog::ZERO );
             }
 
-            if ( !isEditing && le.KeyPress( fheroes2::Key::KEY_DELETE ) && isListboxSelected ) {
-                std::string msg( _( "Are you sure you want to delete file:" ) );
-                msg.append( "\n\n" );
-                msg.append( System::GetBasename( listbox.GetCurrent().file ) );
-                if ( Dialog::YES == fheroes2::showStandardTextMessage( _( "Warning!" ), msg, Dialog::YES | Dialog::NO ) ) {
-                    System::Unlink( listbox.GetCurrent().file );
-                    listbox.RemoveSelected();
-                    if ( lists.empty() || filename.empty() ) {
-                        buttonOk.disable();
-                        isListboxSelected = false;
-                        filename.clear();
-                    }
+            const bool needRedrawListbox = listbox.IsNeedRedraw();
 
-                    listbox.updateScrollBarImage();
-
-                    listbox.SetListContent( lists );
-                }
-
-                needRedraw = true;
+            if ( isEditing && !needRedraw && !isListboxSelected && textInput->eventProcessing() ) {
+                // Text input blinking cursor render is done in Save Game dialog when no file is selected
+                // and when the render of the filename (with cursor) is not planned.
+                display.render( textInput->getCursorArea() );
             }
 
-            // Text input cursor blink.
-            if ( isEditing && Game::validateAnimationDelay( Game::DelayType::CURSOR_BLINK_DELAY ) ) {
-                isCursorVisible = !isCursorVisible;
-                needRedraw = true;
-            }
-
-            if ( !needRedraw && !listbox.IsNeedRedraw() ) {
+            if ( !needRedraw && !needRedrawListbox ) {
                 continue;
             }
 
-            if ( listbox.IsNeedRedraw() ) {
-                listbox.Redraw();
-            }
-
             if ( needRedraw ) {
-                const std::string selectedFileName = isListboxSelected ? ResizeToShortName( listbox.GetCurrent().file ) : "";
-                if ( isListboxSelected && lastSelectedSaveFileName != selectedFileName ) {
-                    lastSelectedSaveFileName = selectedFileName;
-                    filename = selectedFileName;
-                    charInsertPos = filename.size();
+                if ( isListboxSelected ) {
+                    const std::string selectedFileName = System::GetStem( listbox.GetCurrent().filename );
+                    if ( lastSelectedSaveFileName != selectedFileName ) {
+                        lastSelectedSaveFileName = selectedFileName;
+                        filename = selectedFileName;
+                        charInsertPos = filename.size();
+
+                        buttonOkDisabler();
+                    }
+
+                    textInputBackground.restore();
+                    dateBackground.restore();
+
+                    redrawTextInputField( filename, textInputRoi, display );
+                    redrawDateTime( display, listbox.GetCurrent().timestamp, dateTimeoffsetX, textInputRoi.y + 4, fheroes2::FontType::normalYellow() );
                 }
                 else if ( isEditing ) {
                     // Empty last selected save file name so that we can replace the input field's name if we select the same save file again.
                     // But when loading (i.e. isEditing == false), this doesn't matter since we cannot write to the input field
-                    lastSelectedSaveFileName = "";
-                }
+                    lastSelectedSaveFileName.clear();
 
-                textInputAndDateBackground.restore();
-
-                if ( isEditing ) {
-                    isTextLimit = redrawTextInputField( insertCharToString( filename, charInsertPos, isCursorVisible ? '_' : '\x7F' ), textInputRoi, true );
+                    dateBackground.restore();
+                    textInput->draw( filename, static_cast<int32_t>( charInsertPos ) );
                     redrawDateTime( display, std::time( nullptr ), dateTimeoffsetX, textInputRoi.y + 4, fheroes2::FontType::normalWhite() );
                 }
-                else {
-                    redrawTextInputField( filename, textInputRoi, false );
-                    redrawDateTime( display, listbox.GetCurrent().timestamp, dateTimeoffsetX, textInputRoi.y + 4, fheroes2::FontType::normalYellow() );
-                }
             }
 
-            display.render( background.activeArea() );
+            if ( needRedrawListbox ) {
+                listbox.Redraw();
+                display.render( dialogArea );
+            }
+            else {
+                display.render( textInputAndDateROI );
+            }
         }
 
         return result;
