@@ -1,6 +1,6 @@
 /***************************************************************************
  *   fheroes2: https://github.com/ihhub/fheroes2                           *
- *   Copyright (C) 2020 - 2023                                             *
+ *   Copyright (C) 2020 - 2024                                             *
  *                                                                         *
  *   This program is free software; you can redistribute it and/or modify  *
  *   it under the terms of the GNU General Public License as published by  *
@@ -23,6 +23,7 @@
 #include <algorithm>
 #include <cassert>
 #include <cstdint>
+#include <optional>
 #include <string>
 
 #include "agg_image.h"
@@ -42,71 +43,6 @@
 #include "ui_dialog.h"
 #include "ui_text.h"
 
-namespace
-{
-    void changeRaceToNext( Player & player )
-    {
-        switch ( player.GetRace() ) {
-        case Race::KNGT:
-            player.SetRace( Race::BARB );
-            break;
-        case Race::BARB:
-            player.SetRace( Race::SORC );
-            break;
-        case Race::SORC:
-            player.SetRace( Race::WRLK );
-            break;
-        case Race::WRLK:
-            player.SetRace( Race::WZRD );
-            break;
-        case Race::WZRD:
-            player.SetRace( Race::NECR );
-            break;
-        case Race::NECR:
-            player.SetRace( Race::RAND );
-            break;
-        case Race::RAND:
-            player.SetRace( Race::KNGT );
-            break;
-        default:
-            // Did you add a new race? Add the logic above
-            assert( 0 );
-            break;
-        }
-    }
-
-    void changeRaceToPrev( Player & player )
-    {
-        switch ( player.GetRace() ) {
-        case Race::KNGT:
-            player.SetRace( Race::RAND );
-            break;
-        case Race::BARB:
-            player.SetRace( Race::KNGT );
-            break;
-        case Race::SORC:
-            player.SetRace( Race::BARB );
-            break;
-        case Race::WRLK:
-            player.SetRace( Race::SORC );
-            break;
-        case Race::WZRD:
-            player.SetRace( Race::WRLK );
-            break;
-        case Race::NECR:
-            player.SetRace( Race::WZRD );
-            break;
-        case Race::RAND:
-            player.SetRace( Race::NECR );
-            break;
-        default:
-            // Did you add a new race? Add the logic above
-            assert( 0 );
-            break;
-        }
-    }
-}
-
 void Interface::PlayersInfo::UpdateInfo( Players & players, const fheroes2::Point & playerTypeOffset, const fheroes2::Point & classOffset )
 {
     clear();
@@ -139,7 +75,7 @@ void Interface::PlayersInfo::UpdateInfo( Players & players, const fheroes2::Poin
 bool Interface::PlayersInfo::SwapPlayers( Player & player1, Player & player2 ) const
 {
     const Settings & conf = Settings::Get();
-    const Maps::FileInfo & fi = conf.CurrentFileInfo();
+    const Maps::FileInfo & mapInfo = conf.getCurrentMapInfo();
 
     const int player1Color = player1.GetColor();
     const int player2Color = player2.GetColor();
@@ -149,7 +85,7 @@ bool Interface::PlayersInfo::SwapPlayers( Player & player1, Player & player2 ) c
     if ( player1.isControlAI() == player2.isControlAI() ) {
         swap = true;
     }
-    else if ( ( player1Color & fi.AllowCompHumanColors() ) && ( player2Color & fi.AllowCompHumanColors() ) ) {
+    else if ( ( player1Color & mapInfo.AllowCompHumanColors() ) && ( player2Color & mapInfo.AllowCompHumanColors() ) ) {
         const int humans = conf.GetPlayers().GetColors( CONTROL_HUMAN, true );
 
         if ( humans & player1Color ) {
@@ -174,7 +110,7 @@ bool Interface::PlayersInfo::SwapPlayers( Player & player1, Player & player2 ) c
         player2.setHandicapStatus( player1HandicapStatus );
         player1.setHandicapStatus( player2HandicapStatus );
 
-        if ( player1Race != player2Race && conf.AllowChangeRace( player1Color ) && conf.AllowChangeRace( player2Color ) ) {
+        if ( player1Race != player2Race && mapInfo.AllowChangeRace( player1Color ) && mapInfo.AllowChangeRace( player2Color ) ) {
             player1.SetRace( player2Race );
             player2.SetRace( player1Race );
         }
@@ -250,7 +186,7 @@ void Interface::PlayersInfo::RedrawInfo( const bool displayInGameInfo ) const
 {
     const Settings & conf = Settings::Get();
     fheroes2::Display & display = fheroes2::Display::instance();
-    const Maps::FileInfo & fi = conf.CurrentFileInfo();
+    const Maps::FileInfo & mapInfo = conf.getCurrentMapInfo();
 
     const int32_t playerCount = static_cast<int32_t>( conf.GetPlayers().size() );
     const uint32_t humanColors = conf.GetPlayers().GetColors( CONTROL_HUMAN, true );
@@ -264,7 +200,7 @@ void Interface::PlayersInfo::RedrawInfo( const bool displayInGameInfo ) const
             // Current human.
             playerTypeIcnIndex = 9 + Color::GetIndex( info.player->GetColor() );
         }
-        else if ( fi.ComputerOnlyColors() & info.player->GetColor() ) {
+        else if ( mapInfo.ComputerOnlyColors() & info.player->GetColor() ) {
             // Computer only.
             playerTypeIcnIndex = 15 + Color::GetIndex( info.player->GetColor() );
         }
@@ -298,42 +234,9 @@ void Interface::PlayersInfo::RedrawInfo( const bool displayInGameInfo ) const
         name.draw( info.playerTypeRoi.x + 2 + ( maximumTextWidth - name.width() ) / 2, info.playerTypeRoi.y + info.playerTypeRoi.height + 1, display );
 
         // 2. redraw class
-        const bool isActivePlayer = displayInGameInfo ? info.player->isPlay() : conf.AllowChangeRace( info.player->GetColor() );
+        const bool isActivePlayer = displayInGameInfo ? info.player->isPlay() : mapInfo.AllowChangeRace( info.player->GetColor() );
 
-        uint32_t classIcnIndex = 0;
-        switch ( info.player->GetRace() ) {
-        case Race::KNGT:
-            classIcnIndex = isActivePlayer ? 51 : 70;
-            break;
-        case Race::BARB:
-            classIcnIndex = isActivePlayer ? 52 : 71;
-            break;
-        case Race::SORC:
-            classIcnIndex = isActivePlayer ? 53 : 72;
-            break;
-        case Race::WRLK:
-            classIcnIndex = isActivePlayer ? 54 : 73;
-            break;
-        case Race::WZRD:
-            classIcnIndex = isActivePlayer ? 55 : 74;
-            break;
-        case Race::NECR:
-            classIcnIndex = isActivePlayer ? 56 : 75;
-            break;
-        case Race::MULT:
-            classIcnIndex = isActivePlayer ? 57 : 76;
-            break;
-        case Race::RAND:
-            assert( !displayInGameInfo );
-            classIcnIndex = 58;
-            break;
-        default:
-            // Did you add a new race? Add the logic above!
-            assert( 0 );
-            continue;
-        }
-
-        const fheroes2::Sprite & classIcon = fheroes2::AGG::GetICN( ICN::NGEXTRA, classIcnIndex );
+        const fheroes2::Sprite & classIcon = fheroes2::AGG::GetICN( ICN::NGEXTRA, Race::getRaceIcnIndex( info.player->GetRace(), isActivePlayer ) );
         const fheroes2::Sprite & classIconShadow = fheroes2::AGG::GetICN( ICN::NGEXTRA, 61 );
 
         fheroes2::Blit( classIconShadow, display, info.classRoi.x - 5, info.classRoi.y + 3 );
@@ -389,8 +292,8 @@ bool Interface::PlayersInfo::QueueEventProcessing()
     Settings & conf = Settings::Get();
     const LocalEvent & le = LocalEvent::Get();
 
-    if ( le.MousePressRight() ) {
-        const Player * player = GetFromOpponentClick( le.GetMouseCursor() );
+    if ( le.isMouseRightButtonPressed() ) {
+        const Player * player = GetFromOpponentClick( le.getMouseCursorPos() );
         if ( player != nullptr ) {
             fheroes2::showStandardTextMessage(
                 _( "Opponents" ),
@@ -399,7 +302,7 @@ bool Interface::PlayersInfo::QueueEventProcessing()
             return true;
         }
 
-        player = GetFromClassClick( le.GetMouseCursor() );
+        player = GetFromClassClick( le.getMouseCursorPos() );
         if ( player != nullptr ) {
             fheroes2::showStandardTextMessage(
                 _( "Class" ),
@@ -408,26 +311,23 @@ bool Interface::PlayersInfo::QueueEventProcessing()
             return true;
         }
 
-        player = getPlayerFromHandicapRoi( le.GetMouseCursor() );
+        player = getPlayerFromHandicapRoi( le.getMouseCursorPos() );
         if ( player != nullptr ) {
-            fheroes2::showMessage( fheroes2::Text( _( "Handicap" ), fheroes2::FontType::normalYellow() ),
-                                   fheroes2::Text( _( "This lets you change the handicap of a particular player. Only humans may be handicapped. Handicapped players "
-                                                      "start with fewer resources and earn 15 or 30% fewer resources per turn for mild and severe handicaps, "
-                                                      "respectively." ),
-                                                   fheroes2::FontType::normalWhite() ),
-                                   Dialog::ZERO );
+            fheroes2::showStandardTextMessage( _( "Handicap" ),
+                                               _( "This lets you change the handicap of a particular player. Only human players may have a handicap. Handicapped players "
+                                                  "start with fewer resources and earn 15 or 30% fewer resources per turn for mild and severe handicaps, "
+                                                  "respectively." ),
+                                               Dialog::ZERO );
             return true;
         }
 
         return false;
     }
 
-    if ( le.MouseWheelUp() ) {
-        Player * player = GetFromClassClick( le.GetMouseCursor() );
-        if ( player != nullptr ) {
-            if ( conf.AllowChangeRace( player->GetColor() ) ) {
-                changeRaceToPrev( *player );
-            }
+    if ( le.isMouseWheelUp() ) {
+        Player * player = GetFromClassClick( le.getMouseCursorPos() );
+        if ( player != nullptr && conf.getCurrentMapInfo().AllowChangeRace( player->GetColor() ) ) {
+            player->SetRace( Race::getPreviousRace( player->GetRace() ) );
 
             return true;
         }
@@ -435,22 +335,19 @@ bool Interface::PlayersInfo::QueueEventProcessing()
         return false;
     }
 
-    if ( le.MouseWheelDn() ) {
-        Player * player = GetFromClassClick( le.GetMouseCursor() );
-        if ( player != nullptr ) {
-            if ( conf.AllowChangeRace( player->GetColor() ) ) {
-                changeRaceToNext( *player );
-            }
+    if ( le.isMouseWheelDown() ) {
+        Player * player = GetFromClassClick( le.getMouseCursorPos() );
+        if ( player != nullptr && conf.getCurrentMapInfo().AllowChangeRace( player->GetColor() ) ) {
+            player->SetRace( Race::getNextRace( player->GetRace() ) );
 
             return true;
         }
-
         return false;
     }
 
-    Player * player = GetFromOpponentClick( le.GetMouseCursor() );
+    Player * player = GetFromOpponentClick( le.getMouseCursorPos() );
     if ( player != nullptr ) {
-        const Maps::FileInfo & fi = conf.CurrentFileInfo();
+        const Maps::FileInfo & fi = conf.getCurrentMapInfo();
 
         if ( conf.IsGameType( Game::TYPE_MULTI ) ) {
             if ( currentSelectedPlayer == nullptr ) {
@@ -466,7 +363,7 @@ bool Interface::PlayersInfo::QueueEventProcessing()
         else {
             const int playerColor = player->GetColor();
 
-            if ( playerColor & fi.AllowHumanColors() ) {
+            if ( playerColor & fi.colorsAvailableForHumans ) {
                 const int human = conf.GetPlayers().GetColors( CONTROL_HUMAN, true );
 
                 if ( playerColor != human ) {
@@ -487,29 +384,27 @@ bool Interface::PlayersInfo::QueueEventProcessing()
         return true;
     }
 
-    player = GetFromOpponentNameClick( le.GetMouseCursor() );
+    player = GetFromOpponentNameClick( le.getMouseCursorPos() );
     if ( player != nullptr ) {
         std::string str = _( "%{color} player" );
         StringReplace( str, "%{color}", Color::String( player->GetColor() ) );
 
-        std::string res;
-        if ( Dialog::InputString( str, res ) && !res.empty() ) {
+        std::string res = player->GetName();
+        if ( Dialog::inputString( fheroes2::Text{}, fheroes2::Text{ str, fheroes2::FontType::normalWhite() }, res, 0, false, {} ) && !res.empty() ) {
             player->SetName( res );
         }
 
         return true;
     }
 
-    player = GetFromClassClick( le.GetMouseCursor() );
-    if ( player != nullptr ) {
-        if ( conf.AllowChangeRace( player->GetColor() ) ) {
-            changeRaceToNext( *player );
-        }
+    player = GetFromClassClick( le.getMouseCursorPos() );
+    if ( player != nullptr && conf.getCurrentMapInfo().AllowChangeRace( player->GetColor() ) ) {
+        player->SetRace( Race::getNextRace( player->GetRace() ) );
 
         return true;
     }
 
-    player = getPlayerFromHandicapRoi( le.GetMouseCursor() );
+    player = getPlayerFromHandicapRoi( le.getMouseCursorPos() );
     if ( player != nullptr ) {
         if ( !( player->GetControl() & CONTROL_AI ) ) {
             switch ( player->getHandicapStatus() ) {
@@ -537,30 +432,24 @@ bool Interface::PlayersInfo::QueueEventProcessing()
 bool Interface::PlayersInfo::readOnlyEventProcessing()
 {
     const LocalEvent & le = LocalEvent::Get();
-    if ( !le.MousePressRight() ) {
+    if ( !le.isMouseRightButtonPressed() ) {
         // Read only mode works only for right click events.
         return false;
     }
 
-    const Player * player = getPlayerFromHandicapRoi( le.GetMouseCursor() );
+    const Player * player = getPlayerFromHandicapRoi( le.getMouseCursorPos() );
     if ( player != nullptr ) {
         switch ( player->getHandicapStatus() ) {
         case Player::HandicapStatus::NONE:
-            fheroes2::showMessage( fheroes2::Text( _( "No Handicap" ), fheroes2::FontType::normalYellow() ),
-                                   fheroes2::Text( _( "No special restrictions on start resources and earning them per turn." ), fheroes2::FontType::normalWhite() ),
-                                   Dialog::ZERO );
+            fheroes2::showStandardTextMessage( _( "No Handicap" ), _( "No special restrictions on starting resources and resource income per turn." ), Dialog::ZERO );
             break;
         case Player::HandicapStatus::MILD:
-            fheroes2::showMessage( fheroes2::Text( _( "Mild Handicap" ), fheroes2::FontType::normalYellow() ),
-                                   fheroes2::Text( _( "Mild handicapped players start with fewer resources and earn 15% fewer resources per turn." ),
-                                                   fheroes2::FontType::normalWhite() ),
-                                   Dialog::ZERO );
+            fheroes2::showStandardTextMessage( _( "Mild Handicap" ), _( "Players with mild handicap start with fewer resources and earn 15% fewer resources per turn." ),
+                                               Dialog::ZERO );
             break;
         case Player::HandicapStatus::SEVERE:
-            fheroes2::showMessage( fheroes2::Text( _( "Severe Handicap" ), fheroes2::FontType::normalYellow() ),
-                                   fheroes2::Text( _( "Severe handicapped players start with fewer resources and earn 30% fewer resources per turn." ),
-                                                   fheroes2::FontType::normalWhite() ),
-                                   Dialog::ZERO );
+            fheroes2::showStandardTextMessage( _( "Severe Handicap" ),
+                                               _( "Players with severe handicap start with fewer resources and earn 30% fewer resources per turn." ), Dialog::ZERO );
             break;
         default:
             // Did you add a new handicap status? Add the logic above!
