@@ -1,6 +1,6 @@
 /***************************************************************************
  *   fheroes2: https://github.com/ihhub/fheroes2                           *
- *   Copyright (C) 2019 - 2023                                             *
+ *   Copyright (C) 2019 - 2025                                             *
  *                                                                         *
  *   Free Heroes2 Engine: http://sourceforge.net/projects/fheroes2         *
  *   Copyright (C) 2009 by Andrey Afletdinov <fheroes2@gmail.com>          *
@@ -25,13 +25,10 @@
 
 #include <algorithm>
 #include <cassert>
-#include <cstddef>
-#include <memory>
 #include <ostream>
 #include <string>
 #include <vector>
 
-#include "ai.h"
 #include "army.h"
 #include "artifact.h"
 #include "artifact_info.h"
@@ -42,6 +39,7 @@
 #include "difficulty.h"
 #include "game.h"
 #include "game_interface.h"
+#include "game_io.h"
 #include "game_static.h"
 #include "interface_icons.h"
 #include "logging.h"
@@ -50,11 +48,13 @@
 #include "maps_tiles.h"
 #include "maps_tiles_helper.h"
 #include "math_base.h"
+#include "mp2.h"
 #include "payment.h"
 #include "players.h"
 #include "profit.h"
 #include "race.h"
 #include "route.h"
+#include "save_format_version.h"
 #include "serialize.h"
 #include "settings.h"
 #include "skill.h"
@@ -83,7 +83,9 @@ namespace
     Funds getHandicapDependentIncome( const Funds & original, const Player::HandicapStatus handicapStatus )
     {
         const int32_t handicapPercentage = getHandicapIncomePercentage( handicapStatus );
+
         Funds corrected( original );
+
         corrected.wood = std::min( corrected.wood, ( corrected.wood * handicapPercentage + 99 ) / 100 );
         corrected.mercury = std::min( corrected.mercury, ( corrected.mercury * handicapPercentage + 99 ) / 100 );
         corrected.ore = std::min( corrected.ore, ( corrected.ore * handicapPercentage + 99 ) / 100 );
@@ -94,16 +96,15 @@ namespace
 
         return corrected;
     }
-}
 
-bool HeroesStrongestArmy( const Heroes * h1, const Heroes * h2 )
-{
-    return h1 && h2 && h2->GetArmy().isStrongerThan( h1->GetArmy() );
+    bool HeroesStrongestArmy( const Heroes * h1, const Heroes * h2 )
+    {
+        return h1 && h2 && h2->GetArmy().isStrongerThan( h1->GetArmy() );
+    }
 }
 
 Kingdom::Kingdom()
     : color( Color::NONE )
-    , _lastBattleWinHeroID( 0 )
     , lost_town_days( 0 )
     , visited_tents_colors( 0 )
     , _topCastleInKingdomView( -1 )
@@ -112,29 +113,25 @@ Kingdom::Kingdom()
     // Do nothing.
 }
 
-void Kingdom::Init( int clr )
+void Kingdom::Init( const int clr )
 {
     clear();
+
     color = clr;
 
-    if ( Color::ALL & color ) {
-        heroes.reserve( GetMaxHeroes() );
-        castles.reserve( 15 );
+    if ( ( color & Color::ALL ) == 0 ) {
+        DEBUG_LOG( DBG_GAME, DBG_WARN, "Unknown player: " << Color::String( color ) << "(" << static_cast<int>( color ) << ")" )
 
-        // Difficulty calculation is different for campaigns. Difficulty affects only on starting resources for human players.
-        const Settings & configuration = Settings::Get();
-        const int difficultyLevel = ( configuration.isCampaignGameType() ? configuration.CurrentFileInfo().difficulty : configuration.GameDifficulty() );
-
-        resource = _getKingdomStartingResources( difficultyLevel );
-
-        // Some human players can have handicap for resources.
-        const Player * player = Players::Get( color );
-        assert( player != nullptr );
-        resource = getHandicapDependentIncome( resource, player->getHandicapStatus() );
-    }
-    else {
-        DEBUG_LOG( DBG_GAME, DBG_WARN, "Kingdom: unknown player: " << Color::String( color ) << "(" << static_cast<int>( color ) << ")" )
+        return;
     }
+
+    resource = _getKingdomStartingResources( Game::getDifficulty() );
+
+    const Player * player = Players::Get( color );
+    assert( player != nullptr );
+
+    // Some human players can have handicap for resources.
+    resource = getHandicapDependentIncome( resource, player->getHandicapStatus() );
 }
 
 void Kingdom::clear()
@@ -181,25 +178,30 @@ bool Kingdom::isPlay() const
 
 void Kingdom::LossPostActions()
 {
-    if ( isPlay() ) {
-        Players::SetPlayerInGame( color, false );
-
-        // Heroes::Dismiss() calls Kingdom::RemoveHeroes(), which eventually calls heroes.erase()
-        while ( !heroes.empty() ) {
-            Heroes * hero = heroes.back();
-
-            assert( hero->GetColor() == GetColor() );
-
-            hero->Dismiss( static_cast<int>( Battle::RESULT_LOSS ) );
-        }
-
-        if ( !castles.empty() ) {
-            castles.ChangeColors( GetColor(), Color::NONE );
-            castles.clear();
-        }
-
-        world.ResetCapturedObjects( GetColor() );
+    if ( !isPlay() ) {
+        return;
     }
+
+    Players::SetPlayerInGame( color, false );
+
+    // Heroes::Dismiss() calls Kingdom::RemoveHero(), which eventually calls heroes.erase()
+    while ( !heroes.empty() ) {
+        Heroes * hero = heroes.back();
+
+        assert( hero->GetColor() == GetColor() );
+
+        hero->Dismiss( static_cast<int>( Battle::RESULT_LOSS ) );
+    }
+
+    for ( Castle * castle : castles ) {
+        assert( castle != nullptr && castle->GetColor() == GetColor() );
+
+        castle->ChangeColor( Color::NONE );
+    }
+
+    castles.clear();
+
+    world.ResetCapturedObjects( GetColor() );
 }
 
 void Kingdom::ActionBeforeTurn()
@@ -245,19 +247,6 @@ void Kingdom::ActionNewDayResourceUpdate( const std::function<void( const EventD
     // Skip the income for the first day
     if ( world.CountDay() > 1 ) {
         AddFundsResource( GetIncome() );
-
-        // Resource bonuses from campaign awards
-        if ( isControlHuman() && Settings::Get().isCampaignGameType() ) {
-            const std::vector<Campaign::CampaignAwardData> campaignAwards = Campaign::CampaignSaveData::Get().getObtainedCampaignAwards();
-
-            for ( size_t i = 0; i < campaignAwards.size(); ++i ) {
-                if ( campaignAwards[i]._type != Campaign::CampaignAwardData::TYPE_RESOURCE_BONUS ) {
-                    continue;
-                }
-
-                AddFundsResource( Funds( campaignAwards[i]._subType, campaignAwards[i]._amount ) );
-            }
-        }
     }
 
     const bool isAIPlayer = ( GetControl() == CONTROL_AI );
@@ -299,63 +288,57 @@ void Kingdom::ActionNewWeek()
     GetRecruits();
 }
 
-void Kingdom::ActionNewMonth()
+void Kingdom::AddHero( Heroes * hero )
 {
-    // Clear the visited objects with a lifetime of one month, even if this kingdom has already been vanquished
-    visit_object.remove_if( Visit::isMonthLife );
-}
+    if ( hero == nullptr ) {
+        // Why are you adding an empty hero?
+        assert( 0 );
+        return;
+    }
 
-void Kingdom::AddHeroes( Heroes * hero )
-{
-    if ( hero ) {
-        if ( heroes.end() == std::find( heroes.begin(), heroes.end(), hero ) )
-            heroes.push_back( hero );
-
-        const Player * player = Settings::Get().GetPlayers().GetCurrent();
-        if ( player && player->isColor( GetColor() ) && player->isControlHuman() )
-            Interface::AdventureMap::Get().GetIconsPanel().ResetIcons( ICON_HEROES );
-
-        AI::Get().HeroesAdd( *hero );
+    if ( heroes.end() == std::find( heroes.begin(), heroes.end(), hero ) ) {
+        heroes.push_back( hero );
     }
 }
 
-void Kingdom::RemoveHeroes( const Heroes * hero )
+void Kingdom::RemoveHero( const Heroes * hero )
 {
-    if ( hero ) {
-        if ( !heroes.empty() ) {
-            auto it = std::find( heroes.begin(), heroes.end(), hero );
-            assert( it != heroes.end() );
-            if ( it != heroes.end() ) {
-                heroes.erase( it );
-            }
-        }
-
-        Player * player = Players::Get( GetColor() );
-
-        if ( player && player->GetFocus().GetHeroes() == hero ) {
-            player->GetFocus().Reset();
-        }
-
-        assert( hero != nullptr );
-
-        AI::Get().HeroesRemove( *hero );
+    if ( hero == nullptr ) {
+        // Why are you trying to delete a non-existing hero?
+        assert( 0 );
+        return;
     }
 
-    if ( isLoss() )
+    if ( !heroes.empty() ) {
+        auto it = std::find( heroes.begin(), heroes.end(), hero );
+        assert( it != heroes.end() );
+        if ( it != heroes.end() ) {
+            heroes.erase( it );
+        }
+    }
+
+    Player * player = Players::Get( GetColor() );
+
+    if ( player && player->GetFocus().GetHeroes() == hero ) {
+        player->GetFocus().Reset();
+    }
+
+    if ( isLoss() ) {
         LossPostActions();
+    }
 }
 
-void Kingdom::AddCastle( const Castle * castle )
+void Kingdom::AddCastle( Castle * castle )
 {
     if ( castle ) {
-        if ( castles.end() == std::find( castles.begin(), castles.end(), castle ) )
-            castles.push_back( const_cast<Castle *>( castle ) );
+        if ( castles.end() == std::find( castles.begin(), castles.end(), castle ) ) {
+            castles.push_back( castle );
+        }
 
         const Player * player = Settings::Get().GetPlayers().GetCurrent();
-        if ( player && player->isColor( GetColor() ) )
-            Interface::AdventureMap::Get().GetIconsPanel().ResetIcons( ICON_CASTLES );
-
-        AI::Get().CastleAdd( *castle );
+        if ( player && player->isColor( GetColor() ) ) {
+            Interface::AdventureMap::Get().GetIconsPanel().resetIcons( ICON_CASTLES );
+        }
     }
 
     lost_town_days = Game::GetLostTownDays() + 1;
@@ -377,10 +360,6 @@ void Kingdom::RemoveCastle( const Castle * castle )
         if ( player && player->GetFocus().GetCastle() == castle ) {
             player->GetFocus().Reset();
         }
-
-        assert( castle != nullptr );
-
-        AI::Get().CastleRemove( *castle );
     }
 
     if ( isLoss() )
@@ -435,9 +414,9 @@ bool Kingdom::AllowPayment( const Funds & funds ) const
            && ( resource.gold >= funds.gold || 0 == funds.gold );
 }
 
-bool Kingdom::isVisited( const Maps::Tiles & tile ) const
+bool Kingdom::isVisited( const Maps::Tile & tile ) const
 {
-    return isVisited( tile.GetIndex(), tile.GetObject( false ) );
+    return isVisited( tile.GetIndex(), tile.getMainObjectType( false ) );
 }
 
 bool Kingdom::isVisited( int32_t index, const MP2::MapObjectType objectType ) const
@@ -463,24 +442,24 @@ void Kingdom::SetVisited( int32_t index, const MP2::MapObjectType objectType )
         visit_object.emplace_front( index, objectType );
 }
 
-bool Kingdom::isValidKingdomObject( const Maps::Tiles & tile, const MP2::MapObjectType objectType ) const
+bool Kingdom::isValidKingdomObject( const Maps::Tile & tile, const MP2::MapObjectType objectType ) const
 {
-    if ( !MP2::isActionObject( objectType ) )
+    if ( !MP2::isInGameActionObject( objectType ) )
         return false;
 
     if ( isVisited( tile.GetIndex(), objectType ) )
         return false;
 
     // Check castle first to ignore guest hero (tile with both Castle and Hero)
-    if ( tile.GetObject( false ) == MP2::OBJ_CASTLE ) {
+    if ( tile.getMainObjectType( false ) == MP2::OBJ_CASTLE ) {
         const int tileColor = getColorFromTile( tile );
 
         // Castle can only be visited if it either belongs to this kingdom or is an enemy castle (in the latter case, an attack may occur)
         return color == tileColor || !Players::isFriends( color, tileColor );
     }
 
-    // Hero object can overlay other objects when standing on top of it: force check with GetObject( true )
-    if ( objectType == MP2::OBJ_HEROES ) {
+    // Hero object can overlay other objects when standing on top of it: force check with getMainObjectType( true )
+    if ( objectType == MP2::OBJ_HERO ) {
         const Heroes * hero = tile.getHero();
 
         // Hero can only be met if he either belongs to this kingdom or is an enemy hero (in the latter case, an attack will occur)
@@ -538,12 +517,14 @@ uint32_t Kingdom::GetLostTownDays() const
 
 const Recruits & Kingdom::GetRecruits()
 {
-    // In the first week, it is necessary to offer one native hero (or a hero given as a campaign award)
-    const bool offerNativeHero = world.CountWeek() < 2 && recruits.GetID1() == Heroes::UNKNOWN && recruits.GetID2() == Heroes::UNKNOWN;
+    // At the beginning of a new week, in case there are no heroes who retreated or surrendered on the last day of the previous week and should be available for hire
+    // (this mechanic is an extension of fheroes2 and is missing from the original game), it is necessary to offer one native hero
+    const bool offerNativeHero = ( recruits.GetID1() == Heroes::UNKNOWN && recruits.GetID2() == Heroes::UNKNOWN );
     // Special hero given as a campaign award
     Heroes * specialHireableHero = nullptr;
 
-    if ( isControlHuman() && Settings::Get().isCampaignGameType() && offerNativeHero ) {
+    // If there is a hero received as a campaign award, then this hero should be offered instead of a native hero at the beginning of the first week
+    if ( isControlHuman() && Settings::Get().isCampaignGameType() && offerNativeHero && world.CountWeek() < 2 ) {
         const std::vector<Campaign::CampaignAwardData> obtainedAwards = Campaign::CampaignSaveData::Get().getObtainedCampaignAwards();
 
         for ( const auto & obtainedAward : obtainedAwards ) {
@@ -606,24 +587,28 @@ bool Kingdom::AllowRecruitHero( bool check_payment ) const
 
 void Kingdom::ApplyPlayWithStartingHero()
 {
-    if ( !isPlay() || castles.empty() )
+    if ( !isPlay() || castles.empty() ) {
         return;
+    }
 
     bool foundHeroes = false;
 
     for ( const Castle * castle : castles ) {
-        if ( castle == nullptr )
+        if ( castle == nullptr ) {
             continue;
+        }
 
-        // check manual set hero (castle position + point(0, 1))?
+        // Check if there is a hero placed by the map creator near the castle entrance (castle position + point(0, 1))
         const fheroes2::Point & cp = castle->GetCenter();
-        Heroes * hero = world.GetTiles( cp.x, cp.y + 1 ).getHero();
+        Heroes * hero = world.getTile( cp.x, cp.y + 1 ).getHero();
 
-        // and move manual set hero to castle
+        // If there is, move it to the castle
         if ( hero && hero->GetColor() == GetColor() ) {
             const bool patrol = hero->Modes( Heroes::PATROL );
+
             if ( hero->isValid() ) {
                 hero->Move2Dest( Maps::GetIndexFromAbsPoint( cp ) );
+                castle->trainHeroInMageGuild( *hero );
             }
             else {
                 hero->Dismiss( 0 );
@@ -634,19 +619,24 @@ void Kingdom::ApplyPlayWithStartingHero()
                 hero->SetModes( Heroes::PATROL );
                 hero->SetPatrolCenter( cp );
             }
+
             foundHeroes = true;
         }
     }
 
-    if ( !foundHeroes && Settings::Get().GameStartWithHeroes() ) {
-        // get first castle
+    if ( !foundHeroes && Settings::Get().getCurrentMapInfo().startWithHeroInFirstCastle ) {
         const Castle * first = castles.GetFirstCastle();
-        if ( nullptr == first )
+        if ( first == nullptr ) {
             first = castles.front();
+        }
+
+        // If there are no heroes, then there must be at least one castle or town
+        assert( first != nullptr );
 
         Heroes * hero = world.GetHeroForHire( first->GetRace() );
-        if ( hero && AllowRecruitHero( false ) )
+        if ( hero && AllowRecruitHero( false ) ) {
             hero->Recruit( *first );
+        }
     }
 }
 
@@ -655,7 +645,7 @@ uint32_t Kingdom::GetMaxHeroes()
     return GameStatic::GetKingdomMaxHeroes();
 }
 
-Funds Kingdom::GetIncome( int type /* INCOME_ALL */ ) const
+Funds Kingdom::GetIncome( int type /* = INCOME_ALL */ ) const
 {
     Funds totalIncome;
 
@@ -699,11 +689,11 @@ Funds Kingdom::GetIncome( int type /* INCOME_ALL */ ) const
         // estates skill bonus
         for ( const Heroes * hero : heroes ) {
             assert( hero != nullptr );
-            totalIncome.gold += hero->GetSecondaryValues( Skill::Secondary::ESTATES );
+            totalIncome.gold += hero->GetSecondarySkillValue( Skill::Secondary::ESTATES );
         }
     }
 
-    if ( ( type & INCOME_CAMPAIGN_BONUS ) && Settings::Get().isCampaignGameType() ) {
+    if ( isControlHuman() && ( type & INCOME_CAMPAIGN_BONUS ) && Settings::Get().isCampaignGameType() ) {
         const std::vector<Campaign::CampaignAwardData> awards = Campaign::CampaignSaveData::Get().getObtainedCampaignAwards();
         for ( const Campaign::CampaignAwardData & award : awards ) {
             if ( award._type != Campaign::CampaignAwardData::TYPE_RESOURCE_BONUS ) {
@@ -714,15 +704,28 @@ Funds Kingdom::GetIncome( int type /* INCOME_ALL */ ) const
         }
     }
 
-    if ( isControlAI() && totalIncome.gold > 0 ) {
-        const int32_t bonusGold = static_cast<int32_t>( totalIncome.gold * Difficulty::getGoldIncomeBonusForAI( Game::getDifficulty() ) );
+    if ( isControlAI() ) {
+        const Funds incomeBonus = Difficulty::getResourceIncomeBonusForAI( Game::getDifficulty(), *this );
+        if ( incomeBonus.GetValidItemsCount() != 0 ) {
+            DEBUG_LOG( DBG_AI, DBG_TRACE, "AI bonus to the resource income has been applied to " << Color::String( color ) << ": " << incomeBonus.String() )
 
-        totalIncome.gold += bonusGold;
+            totalIncome += incomeBonus;
+        }
+
+        const int32_t goldBonus = static_cast<int32_t>( totalIncome.gold * Difficulty::getGoldIncomeBonusForAI( Game::getDifficulty() ) );
+        if ( goldBonus != 0 ) {
+            DEBUG_LOG( DBG_AI, DBG_TRACE,
+                       "AI bonus to the gold income has been applied to " << Color::String( color ) << ", original income: " << totalIncome.gold
+                                                                          << ", bonus income: " << goldBonus )
+
+            totalIncome.gold += goldBonus;
+        }
     }
 
-    // Some human players can have handicap for resources.
     const Player * player = Players::Get( color );
     assert( player != nullptr );
+
+    // Some human players can have handicap for resources.
     return getHandicapDependentIncome( totalIncome, player->getHandicapStatus() );
 }
 
@@ -768,79 +771,68 @@ double Kingdom::GetArmiesStrength() const
 
 void Kingdoms::Init()
 {
-    const Colors colors( Settings::Get().GetPlayers().GetColors() );
-
     clear();
 
-    for ( Colors::const_iterator it = colors.begin(); it != colors.end(); ++it )
-        GetKingdom( *it ).Init( *it );
+    const Colors colors( Settings::Get().GetPlayers().GetColors() );
+    std::for_each( colors.begin(), colors.end(), [this]( const int color ) { GetKingdom( color ).Init( color ); } );
 }
 
 void Kingdoms::clear()
 {
-    for ( Kingdom & kingdom : kingdoms )
-        kingdom.clear();
+    std::for_each( _kingdoms.begin(), _kingdoms.end(), []( Kingdom & kingdom ) { kingdom.clear(); } );
 }
 
 void Kingdoms::ApplyPlayWithStartingHero()
 {
-    for ( Kingdom & kingdom : kingdoms )
-        if ( kingdom.isPlay() )
+    std::for_each( _kingdoms.begin(), _kingdoms.end(), []( Kingdom & kingdom ) {
+        if ( kingdom.isPlay() ) {
             kingdom.ApplyPlayWithStartingHero();
+        }
+    } );
 }
 
-const Kingdom & Kingdoms::GetKingdom( int color ) const
+const Kingdom & Kingdoms::GetKingdom( const int color ) const
 {
     switch ( color ) {
     case Color::BLUE:
-        return kingdoms[0];
+        return _kingdoms[0];
     case Color::GREEN:
-        return kingdoms[1];
+        return _kingdoms[1];
     case Color::RED:
-        return kingdoms[2];
+        return _kingdoms[2];
     case Color::YELLOW:
-        return kingdoms[3];
+        return _kingdoms[3];
     case Color::ORANGE:
-        return kingdoms[4];
+        return _kingdoms[4];
     case Color::PURPLE:
-        return kingdoms[5];
+        return _kingdoms[5];
     default:
         break;
     }
 
-    return kingdoms[6];
+    return _kingdoms[6];
 }
 
-Kingdom & Kingdoms::GetKingdom( int color )
+Kingdom & Kingdoms::GetKingdom( const int color )
 {
     switch ( color ) {
     case Color::BLUE:
-        return kingdoms[0];
+        return _kingdoms[0];
     case Color::GREEN:
-        return kingdoms[1];
+        return _kingdoms[1];
     case Color::RED:
-        return kingdoms[2];
+        return _kingdoms[2];
     case Color::YELLOW:
-        return kingdoms[3];
+        return _kingdoms[3];
     case Color::ORANGE:
-        return kingdoms[4];
+        return _kingdoms[4];
     case Color::PURPLE:
-        return kingdoms[5];
+        return _kingdoms[5];
     default:
         break;
     }
 
-    return kingdoms[6];
-}
-
-void Kingdom::SetLastBattleWinHero( const Heroes & hero )
-{
-    _lastBattleWinHeroID = hero.GetID();
-}
-
-Heroes * Kingdom::GetLastBattleWinHero() const
-{
-    return Heroes::UNKNOWN != _lastBattleWinHeroID ? world.GetHeroes( _lastBattleWinHeroID ) : nullptr;
+    return _kingdoms[6];
 }
 
 void Kingdom::appendSurrenderedHero( Heroes & hero )
@@ -850,56 +842,56 @@ void Kingdom::appendSurrenderedHero( Heroes & hero )
 
 void Kingdoms::NewDay()
 {
-    for ( Kingdom & kingdom : kingdoms ) {
-        kingdom.ActionNewDay();
-    }
+    std::for_each( _kingdoms.begin(), _kingdoms.end(), []( Kingdom & kingdom ) { kingdom.ActionNewDay(); } );
 }
 
 void Kingdoms::NewWeek()
 {
-    for ( Kingdom & kingdom : kingdoms ) {
-        kingdom.ActionNewWeek();
-    }
-}
-
-void Kingdoms::NewMonth()
-{
-    for ( Kingdom & kingdom : kingdoms ) {
-        kingdom.ActionNewMonth();
-    }
+    std::for_each( _kingdoms.begin(), _kingdoms.end(), []( Kingdom & kingdom ) { kingdom.ActionNewWeek(); } );
 }
 
 int Kingdoms::GetNotLossColors() const
 {
     int result = 0;
-    for ( const Kingdom & kingdom : kingdoms )
-        if ( kingdom.GetColor() && !kingdom.isLoss() )
+    for ( const Kingdom & kingdom : _kingdoms ) {
+        if ( kingdom.GetColor() && !kingdom.isLoss() ) {
             result |= kingdom.GetColor();
+        }
+    }
     return result;
 }
 
-int Kingdoms::FindWins( int cond ) const
+int Kingdoms::FindWins( const int cond ) const
 {
-    for ( const Kingdom & kingdom : kingdoms )
-        if ( kingdom.GetColor() && world.KingdomIsWins( kingdom, cond ) )
+    for ( const Kingdom & kingdom : _kingdoms ) {
+        if ( kingdom.GetColor() && world.KingdomIsWins( kingdom, cond ) ) {
             return kingdom.GetColor();
+        }
+    }
     return 0;
 }
 
 void Kingdoms::AddHeroes( const AllHeroes & heroes )
 {
-    for ( AllHeroes::const_iterator it = heroes.begin(); it != heroes.end(); ++it )
-        // skip gray color
-        if ( ( *it )->GetColor() )
-            GetKingdom( ( *it )->GetColor() ).AddHeroes( *it );
+    for ( Heroes * hero : heroes ) {
+        assert( hero != nullptr );
+
+        // Skip neutral heroes.
+        if ( hero->GetColor() != Color::NONE ) {
+            GetKingdom( hero->GetColor() ).AddHero( hero );
+        }
+    }
 }
 
 void Kingdoms::AddCastles( const AllCastles & castles )
 {
-    for ( const Castle * castle : castles ) {
-        // skip gray color
-        if ( castle->GetColor() )
+    for ( Castle * castle : castles ) {
+        assert( castle != nullptr );
+
+        // Skip neutral castles and towns.
+        if ( castle->GetColor() != Color::NONE ) {
             GetKingdom( castle->GetColor() ).AddCastle( castle );
+        }
     }
 }
 
@@ -907,11 +899,11 @@ std::set<Heroes *> Kingdoms::resetRecruits()
 {
     std::set<Heroes *> remainingRecruits;
 
-    for ( Kingdom & kingdom : kingdoms ) {
+    for ( Kingdom & kingdom : _kingdoms ) {
         Recruits & recruits = kingdom.GetCurrentRecruits();
 
-        // Heroes who surrendered on Sunday should still be available for recruitment next week in
-        // the same kingdom, provided that this kingdom is still playable
+        // Heroes who retreated or surrendered on the last day of the previous week should still be available for recruitment next week in the same kingdom, provided that
+        // this kingdom is still playable. This mechanic is an extension of fheroes2 and is missing from the original game.
         if ( !kingdom.isPlay() || world.CountDay() - recruits.getSurrenderDayOfHero1() > 1 ) {
             recruits.SetHero1( nullptr );
         }
@@ -954,7 +946,7 @@ bool Kingdom::IsTileVisibleFromCrystalBall( const int32_t dest ) const
     return false;
 }
 
-cost_t Kingdom::_getKingdomStartingResources( const int difficulty ) const
+Cost Kingdom::_getKingdomStartingResources( const int difficulty ) const
 {
     if ( isControlAI() ) {
         switch ( difficulty ) {
@@ -994,38 +986,33 @@ cost_t Kingdom::_getKingdomStartingResources( const int difficulty ) const
     return { 7500, 20, 5, 20, 5, 5, 5 };
 }
 
-StreamBase & operator<<( StreamBase & msg, const Kingdom & kingdom )
+OStreamBase & operator<<( OStreamBase & stream, const Kingdom & kingdom )
 {
-    return msg << kingdom.modes << kingdom.color << kingdom.resource << kingdom.lost_town_days << kingdom.castles << kingdom.heroes << kingdom.recruits
-               << kingdom.visit_object << kingdom.puzzle_maps << kingdom.visited_tents_colors << kingdom._lastBattleWinHeroID << kingdom._topCastleInKingdomView
-               << kingdom._topHeroInKingdomView;
+    return stream << kingdom.modes << kingdom.color << kingdom.resource << kingdom.lost_town_days << kingdom.castles << kingdom.heroes << kingdom.recruits
+                  << kingdom.visit_object << kingdom.puzzle_maps << kingdom.visited_tents_colors << kingdom._topCastleInKingdomView << kingdom._topHeroInKingdomView;
 }
 
-StreamBase & operator>>( StreamBase & msg, Kingdom & kingdom )
+IStreamBase & operator>>( IStreamBase & stream, Kingdom & kingdom )
 {
-    return msg >> kingdom.modes >> kingdom.color >> kingdom.resource >> kingdom.lost_town_days >> kingdom.castles >> kingdom.heroes >> kingdom.recruits
-           >> kingdom.visit_object >> kingdom.puzzle_maps >> kingdom.visited_tents_colors >> kingdom._lastBattleWinHeroID >> kingdom._topCastleInKingdomView
-           >> kingdom._topHeroInKingdomView;
-}
+    stream >> kingdom.modes >> kingdom.color >> kingdom.resource >> kingdom.lost_town_days >> kingdom.castles >> kingdom.heroes >> kingdom.recruits
+        >> kingdom.visit_object >> kingdom.puzzle_maps >> kingdom.visited_tents_colors;
 
-StreamBase & operator<<( StreamBase & msg, const Kingdoms & obj )
-{
-    msg << Kingdoms::_size;
-    for ( const Kingdom & kingdom : obj.kingdoms )
-        msg << kingdom;
+    static_assert( LAST_SUPPORTED_FORMAT_VERSION < FORMAT_VERSION_PRE2_1100_RELEASE, "Remove the logic below." );
+    if ( Game::GetVersionOfCurrentSaveFile() < FORMAT_VERSION_PRE2_1100_RELEASE ) {
+        int dummy;
 
-    return msg;
-}
-
-StreamBase & operator>>( StreamBase & msg, Kingdoms & obj )
-{
-    uint32_t kingdomscount = 0;
-    msg >> kingdomscount;
-
-    if ( kingdomscount <= Kingdoms::_size ) {
-        for ( uint32_t i = 0; i < kingdomscount; ++i )
-            msg >> obj.kingdoms[i];
+        stream >> dummy;
     }
 
-    return msg;
+    return stream >> kingdom._topCastleInKingdomView >> kingdom._topHeroInKingdomView;
+}
+
+OStreamBase & operator<<( OStreamBase & stream, const Kingdoms & obj )
+{
+    return stream << obj._kingdoms;
+}
+
+IStreamBase & operator>>( IStreamBase & stream, Kingdoms & obj )
+{
+    return stream >> obj._kingdoms;
 }
